// This class implements some basic functionality to
// lease event registrations.  It can work in 
// conjunction with the BasicUnicastService from
// Chapter 13.
package corejini.chapter14;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMapException;
import net.jini.core.lease.UnknownLeaseException;
import com.sun.jini.lease.landlord.Landlord;
import com.sun.jini.lease.landlord.LandlordLease;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.RemoteException;
import java.util.Vector;
import java.util.Map;
import java.util.HashMap;

public class BasicRegistrationLandlord 
	extends UnicastRemoteObject
    	implements Landlord {
    // A simple leasing policy...10 minute leases.
    protected static final int DEFAULT_MAX_LEASE 
	= 1000 * 60 * 10;
    protected int maxLease = DEFAULT_MAX_LEASE;
    // Assume that registrations are kept in
    // a vector maintained by our clients.
    protected Vector regs;
    // A factory for making landlord leases.
    protected LandlordLease.Factory factory;
    
    public BasicRegistrationLandlord(Vector regs, 
                                     LandlordLease.Factory factory)
        throws RemoteException {
        this.regs = regs;
        this.factory = factory;
    }
   
    // Change the maximum lease time from the default.
    public void setMaxLease(int maxLease) {
        this.maxLease = maxLease;
    }

    // Apply the policy to a requested duration
    // to get an actual expiration time.
    public long getExpiration(long request) {
        if (request > maxLease || request == Lease.ANY)
            return System.currentTimeMillis() + maxLease;
        else
            return System.currentTimeMillis() + request;
    }
   
    // Cancel the lease represented by 'cookie'
    public void cancel(Object cookie) 
        throws UnknownLeaseException {
        for (int i=0, size=regs.size() ; i<size ; i++) {
            Registration reg = 
		(Registration) regs.elementAt(i);
            if (reg.cookie.equals(cookie)) {
                reg.cancelled();
                regs.removeElementAt(i);
                return;
            }
        }
        throw new UnknownLeaseException(cookie.toString());
    }

    // Cancel a set of leases
    public void cancelAll(Object[] cookies) 
        throws LeaseMapException {
        Map exceptionMap = null;
        LeaseMapException lme = null;
        
        for (int i=0 ; i<cookies.length ; i++) {
            try {
                cancel(cookies[i]);
            } catch (UnknownLeaseException ex) {
                if (lme == null) {
                    exceptionMap = new HashMap();
                    lme = new LeaseMapException(
					"cancelAll failed",
                                        exceptionMap);
                }
                Lease lease = factory.newLease(cookies[i],
                                               this, 0);
                exceptionMap.put(lease, ex);
            }
        }
        
        if (lme != null)
            throw lme;
    }

    // Renew the lease specified by 'cookie'
    public long renew(Object cookie, long extension)
        throws UnknownLeaseException {
        
        for (int i=0, size=regs.size() ; i<size ; i++) {
            Registration reg = 
			(Registration) regs.elementAt(i);
            if (reg.getCookie().equals(cookie)) {
                long expiration = 
			getExpiration(extension);
                reg.setExpiration(expiration);
                return expiration - 
			System.currentTimeMillis();
            }
        }
        throw new UnknownLeaseException(cookie.toString());
    }

    // Renew a set of leases.
    public Landlord.RenewResults renewAll(Object[] cookies,
                                          long[] extensions) {
        long[] granted = new long[cookies.length];
        Exception[] denied = null;
        
        for (int i=0 ; i<cookies.length ; i++) {
            try {
                granted[i] = renew(cookies[i], 
				   extensions[i]);
            } catch (Exception ex) {
                if (denied == null) {
                    denied = new Exception[cookies.length+1];
                }
                denied[i+1] = ex;
            }
        }
            
        Landlord.RenewResults results =
            new Landlord.RenewResults(granted, denied);
        return results;
    }
}
