// Create a tunnel between two lookup services

package corejini.chapter8;

import java.io.Serializable;
import java.io.IOException;
import java.util.Hashtable;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import net.jini.lookup.entry.Name;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.core.lookup.ServiceEvent;
import net.jini.core.lookup.ServiceRegistrar;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceMatches;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.EventRegistration;
import com.sun.jini.lookup.JoinManager;
import com.sun.jini.lease.LeaseRenewalManager;
import net.jini.core.discovery.LookupLocator;
import net.jini.discovery.LookupDiscovery;
import net.jini.core.entry.Entry;

class TunnelServiceProxy implements Serializable {
}

public class TunnelService {
    protected JoinManager join;
    protected ServiceTemplate tmpl = 
	new ServiceTemplate(null, null, null);
    protected LeaseRenewalManager mgr = 
	new LeaseRenewalManager();
    protected Hashtable leases = new Hashtable();
    protected Hashtable joinManagers = new Hashtable();
    protected LookupLocator toLoc;
    protected ServiceRegistrar fromReg;
    protected int transitions = 
	ServiceRegistrar.TRANSITION_MATCH_NOMATCH |
        ServiceRegistrar.TRANSITION_NOMATCH_MATCH |
        ServiceRegistrar.TRANSITION_MATCH_MATCH;
    protected Listener listener;

    // Listen for changes in the source lookup service
    class Listener extends UnicastRemoteObject 
		implements RemoteEventListener { 
        public Listener() throws RemoteException {
	}

        public void notify(RemoteEvent ev) throws RemoteException {
            if (!(ev instanceof ServiceEvent)) {
                return;
            }
        
            ServiceEvent sev = (ServiceEvent) ev;
            ServiceItem[] items = new ServiceItem[1];
                items[0] = sev.getServiceItem();
        
            switch (sev.getTransition()) {
            case ServiceRegistrar.TRANSITION_MATCH_NOMATCH:
                withdrawServices(items);
                break;
            case ServiceRegistrar.TRANSITION_NOMATCH_MATCH:
            case ServiceRegistrar.TRANSITION_MATCH_MATCH:
                // new service has shown up, or an old one 
		// has changed--push it to the dest.  if 
		// it's a change we'll overwrite
                pushServices(items);
                break;
            }
        }
    }
    
    public TunnelService() throws RemoteException {
	listener = new Listener();

	if (System.getSecurityManager() == null) {
	    System.setSecurityManager(
		    new RMISecurityManager());
	}
    }
    
    public void createTunnel(String from, 
			     String to) throws Exception {
        LookupLocator fromLoc = new LookupLocator(from);
        toLoc = new LookupLocator(to);
        fromReg = fromLoc.getRegistrar();
        
        publishProxies(fromLoc, toLoc);
        solicitEvents();
        ServiceItem[] fromServices = getServices();
        pushServices(fromServices);
    }
    
    protected ServiceItem[] getServices() throws RemoteException {
        ServiceMatches matches = fromReg.lookup(tmpl, 
						Integer.MAX_VALUE);
        return matches.items;
    }

    protected void publishProxies(LookupLocator from, 
				  LookupLocator to) 
        throws IOException {
        Name name = new Name("Tunnel service");
        ServiceInfo info = new ServiceInfo("Tunnel service",
                                           "Prentice-Hall",
                                           "Your local bookseller",
                                           "v1.0",
                                           null, null);
        Entry[] entries = new Entry[2];
        entries[0] = name;
        entries[1] = info;
        
        LookupLocator[] locators = new LookupLocator[2];
        locators[0] = from;
        locators[1] = to;
        
        String[] groups = LookupDiscovery.NO_GROUPS;
        
        join = new JoinManager(new TunnelServiceProxy(), 
			       entries, groups, 
			       locators, null, mgr);
        
        System.out.println("Join manager started...");
    }
    
    protected void solicitEvents() throws IOException,
            ClassNotFoundException {
        EventRegistration er = fromReg.notify(tmpl, 
					      transitions, 
					      listener,
                                              null, 
					      10 * 60 * 1000);
        
        leases.put(fromReg.getServiceID(), er.getLease());
        mgr.renewUntil(er.getLease(), Long.MAX_VALUE, null);
    }
    
    protected void pushServices(ServiceItem[] items) {
        for (int i=0 ; i<items.length ; i++) {
            try {
                LookupLocator[] toArray = 
			new LookupLocator[] { toLoc };
                System.out.println("Pushing service to " + 
				   toLoc + ": " +
                                   items[i].service);
                JoinManager j = 
		    new JoinManager(items[i].serviceID, 
                                    items[i].service,
                                    items[i].attributeSets,
                                    LookupDiscovery.NO_GROUPS,
                                    toArray, mgr);
                joinManagers.put(items[i].serviceID, j);
            } catch (Exception ex) {
                System.err.println("Couldn't tunnel service: " +
                                   ex.getMessage());
            }
        }
    }
    
    protected void withdrawServices(ServiceItem[] items) {
        for (int i=0 ; i<items.length ; i++) {
            try {
                System.out.println("Withdrawing service " + 
				   items[i].service);
                JoinManager j = (JoinManager) 
                    joinManagers.get(items[i].serviceID);
                j.terminate();
                joinManagers.remove(items[i].serviceID);
            } catch (Exception ex) {
                System.err.println("Couldn't withdraw service: " +
                                   ex.getMessage());
            }
        }
    }
    
    public static void main(String[] args) {
        if (args.length != 2) {
            System.err.println("Usage: TunnelService " +
			       "<from_url> <to_url>");
            System.exit(1);
        }
        
        String from = args[0], to = args[1];
        
        try {
            TunnelService tunnel = new TunnelService();
            tunnel.createTunnel(from, to);
        } catch (Exception ex) {
            System.err.println("Error creating tunnel: " + 
			       ex.getMessage());
            System.exit(1);
        }
    }
}
