// behavior
//
// lookup single click -- update services
// service single click -- update attrs
// attr single click -- noop
//
// lookup double click -- update services
// service double click -- update attrs
// attr double click -- noop

package corejini.chapter9;

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.util.Vector;
import net.jini.admin.*;
import net.jini.discovery.*;
import net.jini.core.lookup.*;
import net.jini.core.event.*;
import net.jini.lookup.DiscoveryAdmin;
import net.jini.lookup.entry.*;
import net.jini.core.entry.Entry;
import com.sun.jini.lease.LeaseRenewalManager;
import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.server.UnicastRemoteObject;
import java.beans.*;
    
// A ListCellRenderer for Lookup ServiceRegistrars --
// draws a ServiceRegistrar as its locator URL
class LookupCellRenderer extends DefaultListCellRenderer {
    public Component getListCellRendererComponent(JList list, 
                                                  Object value,
                                                  int index, 
                                                  boolean isSelected,
                                                  boolean cellHasFocus) {
        Component c = super.getListCellRendererComponent(list, value,
                                                         index, isSelected,
                                                         cellHasFocus);
        if (c instanceof JLabel) {
            JLabel l = (JLabel) c;
            if (value instanceof ServiceRegistrar) {
                ServiceRegistrar reg = (ServiceRegistrar) value;
                try {
                    l.setText(reg.getLocator().toString());
                } catch (RemoteException ex) {
                    System.err.println("Error getting locator: " +
                                       ex.getMessage());
                    l.setText(reg.toString());
                }
            }
        }
        return c;
    }
}
    
// wrap a vector so that we notify listeners when a change occurs
class VectorListModel extends AbstractListModel {
    Vector v = new Vector();
        
    public Object getElementAt(int index) {
        return v.elementAt(index);
    }
    public int getSize() {
        return v.size();
    }
    boolean contains(Object o) {
        return v.contains(o);
    }
    synchronized void addElement(Object o) {
        v.addElement(o);
        fireIntervalAdded(this, v.size(), v.size());
    }
    synchronized void removeElement(Object o) {
        int index = v.indexOf(o);
        if (index != -1) {
            v.remove(index);
            fireIntervalRemoved(this, index, index);
        }
    }
    synchronized void setData(Vector newVector) {
        int len = v.size();
        v = newVector;
        fireIntervalRemoved(this, 0, len);
        fireIntervalAdded(this, 0,  v.size());
    }
}
    
// A ListCellRenderer for services -- draws a service as its name
// if it has one, or a cleaned-up representation of the proxy's
// class
class ServiceCellRenderer extends DefaultListCellRenderer {
    public Component getListCellRendererComponent(JList list, 
                                                  Object value,
                                                  int index, 
                                                  boolean isSelected,
                                                  boolean cellHasFocus) {
        Component c = super.getListCellRendererComponent(list, value,
                                                         index, isSelected,
                                                         cellHasFocus);
        if (c instanceof JLabel) {
            JLabel l = (JLabel) c;
            if (value instanceof ServiceItem) {
                ServiceItem item = (ServiceItem) value;
                String s = null;
                // Use the Name attribute, if there is one.
                for (int i=0 ; i<item.attributeSets.length ; i++) {
                    if (item.attributeSets[i] instanceof Name) {
                        s = ((Name) item.attributeSets[i]).name;
                        break;
                    }
                }
                if (s == null) {
                    s = Browser.declassify(item.service.getClass().getName());
                }
                l.setText(s);
            }
        }
        return c;
    }
}
    
// A ListCellRenderer for attributes -- draws attributes as components
// if they are, or uses the name of the class along with its toString()
// representation
class AttrCellRenderer extends DefaultListCellRenderer {
    public Component getListCellRendererComponent(JList list, Object value,
                                                  int index, 
                                                  boolean isSelected,
                                                  boolean cellHasFocus) {
        Component c = super.getListCellRendererComponent(list, value,
                                                         index, 
                                                         isSelected,
                                                         cellHasFocus);
        return Browser.getComponentForEntry((Entry) value, c);
    }
}
    
// better as an inner class, but iv'e broken it out for legibility.  This class
// handles clicks on list items
class ListClickListener extends MouseAdapter {
    Browser browser;
        
    ListClickListener(Browser browser) {
        this.browser = browser;
    }
    public void mouseClicked(MouseEvent ev) {
        boolean doubleClick = ev.getClickCount() == 2;
            
        if (ev.getSource() instanceof JList) {
            JList list = (JList) ev.getSource();
            int index = list.locationToIndex(ev.getPoint());
                
            if (index == -1)
                return;
                
            if (list.getModel().equals(browser.getLookups())) {
                if (doubleClick) {
                    browser.lookupDoubleClicked(index);
                } else {
                    browser.lookupClicked(index);
                }
            } else if (list.getModel().equals(browser.getServices())) {
                if (doubleClick) {
                    browser.serviceDoubleClicked(index);
                } else {
                    browser.serviceClicked(index);
                }
            }
        }
    }
}

// A class to manage discovery for us
class Discoverer implements DiscoveryListener {
    protected Browser browser;
        
    Discoverer(Browser browser) {
        this.browser = browser;
    }
    public void discovered(DiscoveryEvent ev) {
        ServiceRegistrar regs[] = ev.getRegistrars();
	System.out.println("discovered! " + regs.length);
        for (int i=0  ; i<regs.length ; i++) {
            if (!browser.getLookups().contains(regs[i])) {
		System.out.println("adding " + regs[i]);
                browser.getLookups().addElement(regs[i]);
            } else {
		System.out.println("already contains...");
	    }
        }
    }
    public void discarded(DiscoveryEvent ev) {
        ServiceRegistrar regs[] = ev.getRegistrars();
        for (int i=0  ; i<regs.length ; i++) {
            browser.getLookups().removeElement(regs[i]);
        }
    }
}

// A class to receive RemoteEvents
class Listener extends UnicastRemoteObject implements RemoteEventListener{
    Browser browser;
    
    Listener(Browser browser) throws RemoteException {
        super();
        this.browser = browser;
    }
    public void notify(RemoteEvent ev) {
        if (!(ev instanceof ServiceEvent))
            return;

        // we just update the whole set of services -- not efficient, but
        // shorter than mucking our way through the lists.
        browser.updateServices((ServiceRegistrar) ev.getSource());
    }
}

// This is the overall browser application.  It's a panel, so it can be
// embedded in other JComponents
public class Browser extends JPanel {
    // Models for the three panes
    protected VectorListModel lookups, services, attrs;
    protected ServiceRegistrar currentLookup = null;
    // Search for everything
    protected ServiceTemplate template = 
        new ServiceTemplate(null, new Class[]{ Object.class }, null);
    // Current event registration
    protected EventRegistration er = null;
    protected Listener eventListener;
    protected LeaseRenewalManager leaseMgr = new LeaseRenewalManager();
    protected LookupDiscovery disco;
    
    public Browser() throws IOException, RemoteException { 
        MouseListener listener = new ListClickListener(this);
        
        // Make an event listener
        eventListener = new Listener(this);
        
        // Create the list models
        lookups = new VectorListModel();
        services = new VectorListModel();
        attrs = new VectorListModel();
        
        GridBagLayout layout = new GridBagLayout();
        GridBagConstraints c = new GridBagConstraints();
        
        setLayout(layout);
        
        c.fill = GridBagConstraints.BOTH;
        c.weightx = 1.0;
        c.gridx = GridBagConstraints.RELATIVE;
        c.gridy = 0;
        
        // do labels across the top
        JLabel label = new JLabel("Lookup Services");
        layout.setConstraints(label, c);
        add(label);
        
        label = new JLabel("Services");
        layout.setConstraints(label, c);
        add(label);
        
        label = new JLabel("Attributes");
        layout.setConstraints(label, c);
        add(label);
        
        // do scrolling lists across the bottom
        c.gridy = 1;
        c.weighty = 1.0;
        JList list = new JList(lookups);
        list.addMouseListener(listener);
        list.setCellRenderer(new LookupCellRenderer());
        JScrollPane scroller = new JScrollPane(list);
        layout.setConstraints(scroller, c);
        add(scroller);
        
        list = new JList(services);
        list.setCellRenderer(new ServiceCellRenderer());
        list.addMouseListener(listener);
        scroller = new JScrollPane(list);
        layout.setConstraints(scroller, c);
        add(scroller);
        
        list = new JList(attrs);
        list.setCellRenderer(new AttrCellRenderer());
        list.addMouseListener(listener);
        scroller = new JScrollPane(list);
        layout.setConstraints(scroller, c);
        add(scroller);

        // Set up for discovery
        disco = new LookupDiscovery(LookupDiscovery.ALL_GROUPS);
        disco.addDiscoveryListener(new Discoverer(this));
    }
    
    VectorListModel getLookups() {
        return lookups;
    }
    VectorListModel getServices() {
        return services;
    }
    VectorListModel getAttrs() {
        return attrs;
    }
    
    // Find all services in "reg"
    void updateServices(ServiceRegistrar reg) {
        ServiceMatches matches;
        
        try {
            matches = currentLookup.lookup(template, 
                                           Integer.MAX_VALUE);
        } catch (RemoteException ex) {
            showDialog("Problem contacting registrar:\n" + ex.getMessage());
	    disco.discard(reg);
            return;
        } 
        
        Vector newServices = new Vector(matches.totalMatches);
        for (int i=0 ; i<matches.totalMatches ; i++) {
            newServices.addElement(matches.items[i]);
        }
        
        services.setData(newServices);
    }

    // handle single clicks on lookup services--update the
    // list of services
    protected void lookupClicked(int index) {
        unsolicitEvents();
        currentLookup = (ServiceRegistrar) lookups.getElementAt(index);
        updateServices(currentLookup);
        solicitEvents();
    }
    
    // handle double clicks on lookup services--administer
    // the lookup service
    protected void lookupDoubleClicked(int index) {
        currentLookup = (ServiceRegistrar) lookups.getElementAt(index);
        
        if (!(currentLookup instanceof Administrable)) {
            System.out.println("Lookup service is not administrable");
            return;
        }
        
        Object o;
        
        try {
            o = ((Administrable) currentLookup).getAdmin();
        } catch (RemoteException ex) {
            showDialog("Problem contacting registrar:\n" + ex.getMessage());
	    disco.discard(currentLookup);
            return;
        }
        
        if (o != null) {
            System.err.println("GOT AN ADMINABLE REGISTRAR!");
            AdminPanel panel = new AdminPanel(o);
            JFrame frame = new JFrame("Administration");
            frame.getContentPane().add(panel);
            frame.pack();
            frame.setVisible(true);
        }
    }
    
    // handle single clicks on services--update the attr list
    protected void serviceClicked(int index) {
        ServiceItem item = (ServiceItem) services.getElementAt(index);
        Vector newEntries = new Vector(item.attributeSets.length);
        for (int i=0 ; i<item.attributeSets.length ; i++) {
            newEntries.addElement(item.attributeSets[i]);
        }
        
        attrs.setData(newEntries);
    }

    // handle double clicks on services--try to administer
    // them
    protected void serviceDoubleClicked(int index) {
        ServiceItem item = (ServiceItem) services.getElementAt(index);
        
        if (!(item.service instanceof Administrable)) {
            System.out.println("Service is not administrable");
            return;
        }
        
        Object o;
        
        try {
            o = ((Administrable) item.service).getAdmin();
        } catch (RemoteException ex) {
            showDialog("Problem contacting registrar:\n" + ex.getMessage());
            return;
        }
        
        if (o != null) {
            AdminPanel panel = new AdminPanel(o);
            JFrame frame = new JFrame("Administration");
            frame.getContentPane().add(panel);
            frame.pack();
            frame.setVisible(true);
        }
        
        System.out.println("Got admin object: " + o);
    }
    
    // Ask a lookup service for events
    void solicitEvents() {
        System.out.println("Soliciting events");
        try {
            er = currentLookup.notify(template, 
                                  ServiceRegistrar.TRANSITION_MATCH_NOMATCH |
                                  ServiceRegistrar.TRANSITION_NOMATCH_MATCH |
                                  ServiceRegistrar.TRANSITION_MATCH_MATCH,
                                  eventListener, null, 10 * 60 * 1000);
        
            leaseMgr.renewUntil(er.getLease(), Long.MAX_VALUE, null);
        } catch (RemoteException ex) {
            er = null;
	    disco.discard(currentLookup);
        }
    }
    
    // Stop receiving events
    void unsolicitEvents() {
        if (er == null)
            return;
        
        System.out.println("UN-Soliciting events");
        try {
            leaseMgr.cancel(er.getLease());
        } catch (Exception ex) {
        }
    }
    
    // Clean up a classname (strip off package part)
    static String declassify(String name) {
        if (name == null) return null;
        
        int idx = name.lastIndexOf('.');
        
        if (idx == -1)
            return name;
        else
            return name.substring(idx+1);
    }
    
    void showDialog(String message) {
        JOptionPane pane = new JOptionPane(message, JOptionPane.DEFAULT_OPTION,
                                           JOptionPane.ERROR_MESSAGE,
                                           null, null, null);
        JDialog dialog = new JDialog(JOptionPane.getFrameForComponent(this), 
                                     "Alert!");
        dialog.getContentPane().setLayout(new BorderLayout());
        dialog.getContentPane().add(pane, BorderLayout.CENTER);
        dialog.pack();
        dialog.show();
    }
    
    static Component getComponentForEntry(Entry e, Component defaultComponent) {
        EntryBean bean = null;

	// If we're passed a null entry, return the
	// default component (if it's a label, make 
	// it say "NULL")
	if (e == null) {
	    if (defaultComponent instanceof JLabel) {
	        ((JLabel) defaultComponent).setText("NULL");
            }
	    return defaultComponent;
        }
        
        try {
            bean = EntryBeans.createBean(e);
        } catch (Exception ex) {
        }
                    
        // Now beanInstance will be either null or will ref a bean.
        // If it's null or it's not a component, we just tostring it.  If
        // it's a component, we return that.
        if (bean == null || !(bean instanceof Component)) {
            if (defaultComponent instanceof JLabel) {
                String cname = 
                    Browser.declassify(e.getClass().getName());
                ((JLabel) defaultComponent).setText(cname + " [" + 
                                                    e.toString() + "]");
            }
            return defaultComponent;
        } else {
            ((Component) bean).setVisible(true);
            return (Component) bean;
        }
    }
    
    public static void main(String[] args) {
        try {
	    if (System.getSecurityManager() == null) {
		System.setSecurityManager(new RMISecurityManager());
	    }
		
            JFrame frame = new JFrame("Browser");
            frame.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent ev) {
                    System.exit(1);
                }
            });
            frame.getContentPane().add(new Browser());
            frame.pack();
            frame.setVisible(true);
        } catch (Exception ex) {
            System.err.println("Couldn't create browser:" + ex.getMessage());
            ex.printStackTrace();
        }
    }
}
